/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdlib.h>
#include <gtest/gtest.h>
#include <app_control_uri.h>

#include <iostream>
#include <memory>

#include "uri_internal.h"

// The source of the following example is "https://en.wikipedia.org/wiki/Uniform_Resource_Identifier"
static const char EXAM_URI1[] = "https://john.doe@www.example.com:123/forum/questions/?tag=networking&order=newest#top";
static const char EXAM_URI2[] = "ldap://[2001:db8::7]/c=GB?objectClass?one";
static const char EXAM_URI3[] = "mailto:John.Doe@example.com";
static const char EXAM_URI4[] = "news:comp.infosystems.www.servers.unix";
static const char EXAM_URI5[] = "tel:+1-816-555-1212";
static const char EXAM_URI6[] = "telnet://192.0.2.16:80/";
static const char EXAM_URI7[] = "urn:oasis:names:specification:docbook:dtd:xml:4.1.2";

static const char ENCODED_URI1[] = "https:%2F%2Fjohn.doe%40www.example.com%3A123%2Fforum%2Fquestions%2F?tag=networking&order=newest#top";
static const char ENCODED_URI2[] = "ldap:%2F%2F%5B2001%3Adb8%3A%3A7%5D%2Fc%3DGB?objectClass?one";
static const char ENCODED_URI3[] = "mailto:John.Doe%40example.com";
static const char ENCODED_URI4[] = "news:comp.infosystems.www.servers.unix";
static const char ENCODED_URI5[] = "tel:%2B1-816-555-1212";
static const char ENCODED_URI6[] = "telnet:%2F%2F192.0.2.16%3A80%2F";
static const char ENCODED_URI7[] = "urn:oasis%3Anames%3Aspecification%3Adocbook%3Adtd%3Axml%3A4.1.2";

extern int gargc;
extern char** gargv;

class UriTest : public testing::Test {
 public:
  virtual ~UriTest() {
    if (builder_ != nullptr)
      app_control_uri_builder_destroy(builder_);
    if (uri_ != nullptr)
      app_control_uri_destroy(uri_);
  }

  virtual void SetUp() {
    std::string exam_uri[] = {
      ENCODED_URI1,
      ENCODED_URI2,
      ENCODED_URI3,
      ENCODED_URI4,
      ENCODED_URI5,
      ENCODED_URI6,
      ENCODED_URI7,
    };

    for (int i = 0; i < 7; i++) {
      uri_list_[i] =
        std::unique_ptr<appcontrol::Uri>(new appcontrol::Uri(exam_uri[i]));
    }

    app_control_uri_builder_create(&builder_);
    EXPECT_NE(builder_, nullptr);

    app_control_uri_builder_set_scheme(builder_, "https");
    app_control_uri_builder_set_authority(builder_,
        "//john.doe@www.example.com:123");
    app_control_uri_builder_set_path(builder_, "/forum/questions/");
    app_control_uri_builder_set_fragment(builder_, "top");
    app_control_uri_builder_add_query(builder_, "tag", "networking");
    app_control_uri_builder_add_query(builder_, "order", "newest");

    app_control_uri_builder_build(builder_, &uri_);
    EXPECT_NE(uri_, nullptr);
  }

  virtual void TearDown() {
    count_ = 0;

    if (builder_ != nullptr) {
      auto builder =
        std::unique_ptr<void, decltype(app_control_uri_builder_destroy)*>(
          builder_, app_control_uri_builder_destroy);
      builder_ = nullptr;
    }
    if (uri_ != nullptr) {
      auto uri = std::unique_ptr<void, decltype(app_control_uri_destroy)*>(uri_,
          app_control_uri_destroy);
      uri_ = nullptr;
    }
  }

  std::unique_ptr<appcontrol::Uri> uri_list_[7];
  app_control_uri_builder_h builder_;
  app_control_uri_h uri_;
  int count_ = 0;
};

/*
 * @testcase Uri_Constructor
 * @description Creates a Uri object.
 * @apicovered appcontrol::Uri
 */
TEST_F(UriTest, Uri_Constructor) {
  std::unique_ptr<appcontrol::Uri> uri = std::unique_ptr<appcontrol::Uri>(
      new appcontrol::Uri(ENCODED_URI1));
  EXPECT_NE(uri, nullptr);
}

/*
 * @testcase Uri_GetScheme
 * @description Gets a scheme component from a URI object.
 * @apicovered appcontrol::Uri::GetScheme
 */
TEST_F(UriTest, Uri_GetScheme) {
  std::string exam_scheme[] = {
    "https",
    "ldap",
    "mailto",
    "news",
    "tel",
    "telnet",
    "urn",
  };

  for (int i = 0; i < 7; i++) {
    EXPECT_EQ(uri_list_[i]->GetScheme(), exam_scheme[i]);
  }
}

/*
 * @testcase Uri_GetAuthority
 * @description Gets an authority component from a URI object.
 * @apicovered appcontrol::Uri::GetAuthority
 */
TEST_F(UriTest, Uri_GetAuthority) {
  std::string exam_auth[] = {
    "//john.doe@www.example.com:123",
    "//[2001:db8::7]",
    "",
    "",
    "",
    "//192.0.2.16:80",
    "",
  };

  for (int i = 0; i < 7; i++) {
    EXPECT_EQ(uri_list_[i]->GetAuthority(), exam_auth[i]);
  }
}

/*
 * @testcase Uri_GetPath
 * @description Gets a path component from a URI object.
 * @apicovered appcontrol::Uri::GetPath
 */
TEST_F(UriTest, Uri_GetPath) {
  std::string exam_path[] = {
    "/forum/questions/",
    "/c=GB",
    "John.Doe@example.com",
    "comp.infosystems.www.servers.unix",
    "+1-816-555-1212",
    "/",
    "oasis:names:specification:docbook:dtd:xml:4.1.2",
  };

  for (int i = 0; i < 7; i++) {
    EXPECT_EQ(uri_list_[i]->GetPath(), exam_path[i]);
  }
}

/*
 * @testcase Uri_GetFragment
 * @description Gets a fragment component from a URI object.
 * @apicovered appcontrol::Uri::GetFragment
 */
TEST_F(UriTest, Uri_GetFragment) {
  std::string exam_fragment[] = {
    "top",
    "",
    "",
    "",
    "",
    "",
    "",
  };

  for (int i = 0; i < 7; i++) {
    EXPECT_EQ(uri_list_[i]->GetFragment(), exam_fragment[i]);
  }
}

/*
 * @testcase Uri_GetHost
 * @description Gets a host subcomponent from a URI object.
 * @apicovered appcontrol::Uri::GetHost
 */
TEST_F(UriTest, Uri_GetHost) {
  std::string exam_host[] = {
    "www.example.com",
    "[2001:db8::7]",
    "",
    "",
    "",
    "192.0.2.16",
    "",
  };

  for (int i = 0; i < 7; i++) {
    EXPECT_EQ(uri_list_[i]->GetHost(), exam_host[i]);
  }
}

/*
 * @testcase Uri_GetPort
 * @description Gets a port subcomponent from a URI object.
 * @apicovered appcontrol::Uri::GetPort
 */
TEST_F(UriTest, Uri_GetPort) {
  std::string exam_port[] = {
    "123",
    "",
    "",
    "",
    "",
    "80",
    "",
  };

  for (int i = 0; i < 7; i++) {
    EXPECT_EQ(uri_list_[i]->GetPort(), exam_port[i]);
  }
}

/*
 * @testcase Uri_GetUser
 * @description Gets a user subcomponent from a URI object.
 * @apicovered appcontrol::Uri::GetUser
 */
TEST_F(UriTest, Uri_GetUser) {
  std::string exam_user[] = {
    "john.doe",
    "",
    "",
    "",
    "",
    "",
    "",
  };

  for (int i = 0; i < 7; i++) {
    EXPECT_EQ(uri_list_[i]->GetUser(), exam_user[i]);
  }
}

/*
 * @testcase Uri_GetQuery
 * @description Gets a query component from a URI object.
 * @apicovered appcontrol::Uri::GetQuery
 */
TEST_F(UriTest, Uri_GetQuery) {
  int match_count = 0;
  auto& query = uri_list_[0]->GetQuery();
  for (auto& i : query) {
    if (i.first == "tag" &&
        i.second == "networking") {
      match_count++;
    } else if (i.first == "order" &&
               i.second == "newest") {
      match_count++;
    }
  }

  EXPECT_EQ(match_count, 2);
}

/*
 * @testcase Uri_Encode
 * @description Encodes a URI object to string.
 * @apicovered appcontrol::Uri::Encode
 */
TEST_F(UriTest, Uri_Encode) {
  std::unique_ptr<appcontrol::Uri> uri = std::unique_ptr<appcontrol::Uri>(
      new appcontrol::Uri(EXAM_URI1));
  EXPECT_NE(uri, nullptr);
  EXPECT_EQ(uri->Encode(), ENCODED_URI1);
}

/*
 * @testcase Uri_Builder
 * @description Builds a URI builder object to create a URI object.
 * @apicovered appcontrol::Uri::Builder, appcontrol::Uri::Builder::SetScheme,
 *             appcontrol::Uri::Builder::SetAuthority, appcontrol::Uri::Builder::SetPath,
 *             appcontrol::Uri::Builder::AddQuery, appcontrol::Uri::Builder::SetFragment,
 *             appcontrol::Uri::Builder::Build
 */
TEST_F(UriTest, Uri_Builder) {
  std::unique_ptr<appcontrol::Uri::Builder> builder(new appcontrol::Uri::Builder());
  EXPECT_NE(builder, nullptr);
  builder->SetScheme("https");
  builder->SetAuthority("//john.doe@www.example.com:123");
  builder->SetPath("/forum/questions/");
  builder->AddQuery("tag", "networking");
  builder->AddQuery("order", "newest");
  builder->SetFragment("top");

  appcontrol::Uri uri = builder->Build();
  EXPECT_EQ(uri.Encode(), ENCODED_URI1);
}

/*
 * @testcase app_control_uri_builder_create_p
 * @description Creates a URI builder handle.
 * @apicovered app_control_uri_builder_create
 */
TEST_F(UriTest, app_control_uri_builder_create_p) {
  app_control_uri_builder_h builder = nullptr;
  int r = app_control_uri_builder_create(&builder);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  EXPECT_NE(builder, nullptr);
  app_control_uri_builder_destroy(builder);
}

/*
 * @testcase app_control_uri_builder_create_n
 * @description Creates a URI builder handle.
 *              The function returns a negative error value.
 * @apicovered app_control_uri_builder_create
 */
TEST_F(UriTest, app_control_uri_builder_create_n) {
  int r = app_control_uri_builder_create(nullptr);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_builder_set_scheme_p
 * @description Sets a scheme component to a URI builder handle.
 * @apicovered app_control_uri_builder_set_scheme
 */
TEST_F(UriTest, app_control_uri_builder_set_scheme_p) {
  int r = app_control_uri_builder_set_scheme(builder_, "https");
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
}

/*
 * @testcase app_control_uri_builder_set_scheme_n
 * @description Sets a scheme component to a URI builder handle.
 *              The function returns a negative error value.
 * @apicovered app_control_uri_builder_set_scheme
 */
TEST_F(UriTest, app_control_uri_builder_set_scheme_n) {
  int r = app_control_uri_builder_set_scheme(builder_, nullptr);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);

  r = app_control_uri_builder_set_scheme(nullptr, "https");
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_builder_set_authority_p
 * @description Sets an authority component to a URI builder handle.
 * @apicovered app_control_uri_builder_set_authority
 */
TEST_F(UriTest, app_control_uri_builder_set_authority_p) {
  int r = app_control_uri_builder_set_authority(builder_,
      "//john.doe@www.example.com:123");
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
}

/*
 * @testcase app_control_uri_builder_set_authority_n
 * @description Sets an authority component to a URI builder handle.
 *              The function returns a negative error value.
 * @apicovered app_control_uri_builder_set_authority
 */
TEST_F(UriTest, app_control_uri_builder_set_authority_n) {
  int r = app_control_uri_builder_set_authority(builder_, nullptr);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);

  r = app_control_uri_builder_set_authority(nullptr,
      "//john.doe@www.example.com:123");
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_builder_set_path_p
 * @description Sets a path component to a URI builder handle.
 * @apicovered app_control_uri_builder_set_path
 */
TEST_F(UriTest, app_control_uri_builder_set_path_p) {
  int r = app_control_uri_builder_set_path(builder_, "/forum/questions/");
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
}

/*
 * @testcase app_control_uri_builder_set_path_n
 * @description Sets a path component to a URI builder handle.
 *              The function returns a negative error value.
 * @apicovered app_control_uri_builder_set_path
 */
TEST_F(UriTest, app_control_uri_builder_set_path_n) {
  int r = app_control_uri_builder_set_path(builder_, nullptr);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);

  r = app_control_uri_builder_set_path(nullptr, "/forum/questions/");
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_builder_add_path_p
 * @description Adds a path component on a URI builder handle.
 * @apicovered app_control_uri_builder_add_path
 */
TEST_F(UriTest, app_control_uri_builder_add_path_p) {
  int r = app_control_uri_builder_add_path(builder_, "/forum");
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);

  r = app_control_uri_builder_add_path(builder_, "/questions/");
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
}

/*
 * @testcase app_control_uri_builder_add_path_n
 * @description Adds a path component on a URI builder handle.
 *              The function returns a negative error value.
 * @apicovered app_control_uri_builder_add_path
 */
TEST_F(UriTest, app_control_uri_builder_add_path_n) {
  int r = app_control_uri_builder_add_path(builder_, nullptr);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);

  r = app_control_uri_builder_add_path(nullptr, "/questions/");
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_builder_set_fragment_p
 * @description Sets a fragment component to a URI builder handle.
 * @apicovered app_control_uri_builder_set_fragment
 */
TEST_F(UriTest, app_control_uri_builder_set_fragment_p) {
  int r = app_control_uri_builder_set_fragment(builder_, "top");
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
}

/*
 * @testcase app_control_uri_builder_set_fragment_n
 * @description Sets a fragment component to a URI builder handle.
 *              The function returns a negative error value.
 * @apicovered app_control_uri_builder_set_fragment
 */
TEST_F(UriTest, app_control_uri_builder_set_fragment_n) {
  int r = app_control_uri_builder_set_fragment(builder_, nullptr);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);

  r = app_control_uri_builder_set_fragment(nullptr, "top");
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_builder_add_query_p
 * @description Adds a query component on a URI builder handle.
 * @apicovered app_control_uri_builder_add_query
 */
TEST_F(UriTest, app_control_uri_builder_add_query_p) {
  int r = app_control_uri_builder_add_query(builder_, "tag", "networking");
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);

  r = app_control_uri_builder_add_query(builder_, "order", "newest");
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
}

/*
 * @testcase app_control_uri_builder_add_query_n
 * @description Adds a query component on a URI builder handle.
 *              The function returns a negative error value.
 * @apicovered app_control_uri_builder_add_query
 */
TEST_F(UriTest, app_control_uri_builder_add_query_n) {
  int r = app_control_uri_builder_add_query(builder_, "tag", nullptr);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);

  r = app_control_uri_builder_add_query(builder_, nullptr, "newest");
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);

  r = app_control_uri_builder_add_query(nullptr, "tag", "newest");
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_builder_build_p
 * @description Builds a URI builder handle to create a URI handle.
 * @apicovered app_control_uri_builder_build
 */
TEST_F(UriTest, app_control_uri_builder_build_p) {
  app_control_uri_h app_control_uri = nullptr;
  int r = app_control_uri_builder_build(builder_, &app_control_uri);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  EXPECT_NE(app_control_uri, nullptr);
  app_control_uri_destroy(app_control_uri);
}

/*
 * @testcase app_control_uri_builder_build_n
 * @description Builds a URI builder handle to create a URI handle.
 *              The function returns a negative error value.
 * @apicovered app_control_uri_builder_build
 */
TEST_F(UriTest, app_control_uri_builder_build_n) {
  int r = app_control_uri_builder_build(builder_, nullptr);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);

  app_control_uri_h app_control_uri = nullptr;
  r = app_control_uri_builder_build(nullptr, &app_control_uri);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_builder_destroy_p
 * @description Destroys a URI builder handle.
 * @apicovered app_control_uri_builder_destroy
 */
TEST_F(UriTest, app_control_uri_builder_destroy_p) {
  int r = app_control_uri_builder_destroy(builder_);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  builder_ = nullptr;
}

/*
 * @testcase app_control_uri_builder_destroy_n
 * @description Destroys a URI builder handle.
 *              The function returns a negative error value.
 * @apicovered app_control_uri_builder_destroy
 */
TEST_F(UriTest, app_control_uri_builder_destroy_n) {
  int r = app_control_uri_builder_destroy(nullptr);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_create_p
 * @description Creates a URI builder handle from an encoded URI string.
 * @apicovered app_control_uri_create
 */
TEST_F(UriTest, app_control_uri_create_p) {
  app_control_uri_h app_control_uri = nullptr;
  const char *scheme;
  const char *authority;
  const char *path;
  const char *fragment;
  const char *host;
  const char *port;
  const char *user;
  int r = app_control_uri_create(ENCODED_URI1, &app_control_uri);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);

  auto handle = std::unique_ptr<void, decltype(app_control_uri_destroy)*>(
      app_control_uri, app_control_uri_destroy);

  r = app_control_uri_get_scheme(handle.get(), &scheme);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  r = app_control_uri_get_authority(handle.get(), &authority);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  r = app_control_uri_get_path(handle.get(), &path);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  r = app_control_uri_get_fragment(handle.get(), &fragment);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);

  EXPECT_EQ(std::string(scheme), "https");
  EXPECT_EQ(std::string(authority), "//john.doe@www.example.com:123");
  EXPECT_EQ(std::string(path), "/forum/questions/");
  EXPECT_EQ(std::string(fragment), "top");

  app_control_uri_query_h query = nullptr;
  r = app_control_uri_get_query(handle.get(), &query);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  EXPECT_NE(query, nullptr);

  r = app_control_uri_get_host(handle.get(), &host);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);

  r = app_control_uri_get_port(handle.get(), &port);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);

  r = app_control_uri_get_user(handle.get(), &user);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);

  EXPECT_EQ(std::string(host), "www.example.com");
  EXPECT_EQ(std::string(port), "123");
  EXPECT_EQ(std::string(user), "john.doe");
}

/*
 * @testcase app_control_uri_create_n
 * @description Creates a URI builder handle from an encoded URI string.
 *              The function returns a negative error value.
 * @apicovered app_control_uri_create
 */
TEST_F(UriTest, app_control_uri_create_n) {
  int r = app_control_uri_create(ENCODED_URI1, nullptr);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);

  app_control_uri_h app_control_uri = nullptr;
  r = app_control_uri_create(nullptr, &app_control_uri);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_destroy_p
 * @description Destroys a URI handle.
 * @apicovered app_control_uri_destroy
 */
TEST_F(UriTest, app_control_uri_destroy_p) {
  int r = app_control_uri_destroy(uri_);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  uri_ = nullptr;
}

/*
 * @testcase app_control_uri_destroy_n
 * @description Destroys a URI handle.
 *              The function returns a negative error value.
 * @apicovered app_control_uri_destroy
 */
TEST_F(UriTest, app_control_uri_destroy_n) {
  int r = app_control_uri_destroy(nullptr);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_encode_p
 * @description Encodes a URI handle to string.
 * @apicovered app_control_uri_encode
 */
TEST_F(UriTest, app_control_uri_encode_p) {
  char *encoded_app_control_uri = nullptr;
  int r = app_control_uri_encode(uri_, &encoded_app_control_uri);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  EXPECT_EQ(std::string(encoded_app_control_uri), ENCODED_URI1);
  if (encoded_app_control_uri)
    free(encoded_app_control_uri);
}

/*
 * @testcase app_control_uri_encode_n
 * @description Encodes a URI handle to string.
 *              The function returns a nullptr.
 * @apicovered app_control_uri_encode
 */
TEST_F(UriTest, app_control_uri_encode_n) {
  char *encoded_app_control_uri;
  int r = app_control_uri_encode(nullptr, &encoded_app_control_uri);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_get_scheme_p
 * @description Gets a scheme component from a URI handle.
 * @apicovered app_control_uri_get_scheme
 */
TEST_F(UriTest, app_control_uri_get_scheme_p) {
  const char* scheme;
  int r = app_control_uri_get_scheme(uri_, &scheme);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  EXPECT_NE(scheme, nullptr);
  EXPECT_EQ(std::string(scheme), "https");
}

/*
 * @testcase app_control_uri_get_scheme_n
 * @description Gets a scheme component from a URI handle.
 *              The function returns a nullptr.
 * @apicovered app_control_uri_get_scheme
 */
TEST_F(UriTest, app_control_uri_get_scheme_n) {
  const char* scheme;
  int r = app_control_uri_get_scheme(nullptr, &scheme);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_get_authority_p
 * @description Gets an authority component from a URI handle.
 * @apicovered app_control_uri_get_authority
 */
TEST_F(UriTest, app_control_uri_get_authority_p) {
  const char* auth;
  int r = app_control_uri_get_authority(uri_, &auth);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  EXPECT_NE(auth, nullptr);
  EXPECT_EQ(std::string(auth), "//john.doe@www.example.com:123");
}

/*
 * @testcase app_control_uri_get_authority_n
 * @description Gets an authority component from a URI handle.
 *              The function returns a nullptr.
 * @apicovered app_control_uri_get_authority
 */
TEST_F(UriTest, app_control_uri_get_authority_n) {
  const char* auth;
  int r = app_control_uri_get_authority(nullptr, &auth);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_get_path_p
 * @description Gets a path component from a URI handle.
 * @apicovered app_control_uri_get_path
 */
TEST_F(UriTest, app_control_uri_get_path_p) {
  const char* path;
  int r = app_control_uri_get_path(uri_, &path);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  EXPECT_NE(path, nullptr);
  EXPECT_EQ(std::string(path), "/forum/questions/");
}

/*
 * @testcase app_control_uri_get_path_n
 * @description Gets a path component from a URI handle.
 *              The function returns a nullptr.
 * @apicovered app_control_uri_get_path
 */
TEST_F(UriTest, app_control_uri_get_path_n) {
  const char* path;
  int r = app_control_uri_get_path(nullptr, &path);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_get_fragment_p
 * @description Gets a fragment component from a URI handle.
 * @apicovered app_control_uri_get_fragment
 */
TEST_F(UriTest, app_control_uri_get_fragment_p) {
  const char* fragment;
  int r = app_control_uri_get_fragment(uri_, &fragment);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  EXPECT_NE(fragment, nullptr);
  EXPECT_EQ(std::string(fragment), "top");
}

/*
 * @testcase app_control_uri_get_fragment_n
 * @description Gets a fragment component from a URI handle.
 *              The function returns a nullptr.
 * @apicovered app_control_uri_get_fragment
 */
TEST_F(UriTest, app_control_uri_get_fragment_n) {
  const char* fragment;
  int r = app_control_uri_get_fragment(nullptr, &fragment);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_get_query_p
 * @description Gets a handle of the query component from a URI handle.
 * @apicovered app_control_uri_get_query
 */
TEST_F(UriTest, app_control_uri_get_query_p) {
  app_control_uri_query_h query = nullptr;
  int r = app_control_uri_get_query(uri_, &query);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  EXPECT_NE(query, nullptr);
}

/*
 * @testcase app_control_uri_get_query_n
 * @description Gets a handle of the query component from a URI handle.
 *              The function returns a negative error value.
 * @apicovered app_control_uri_get_query
 */
TEST_F(UriTest, app_control_uri_get_query_n) {
  int r = app_control_uri_get_query(uri_, nullptr);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);

  app_control_uri_query_h query = nullptr;
  r = app_control_uri_get_query(nullptr, &query);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_get_host_p
 * @description Gets a host subcomponent from a URI handle.
 * @apicovered app_control_uri_get_host
 */
TEST_F(UriTest, app_control_uri_get_host_p) {
  const char* host;
  int r = app_control_uri_get_host(uri_, &host);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  EXPECT_NE(host, nullptr);
  EXPECT_EQ(std::string(host), "www.example.com");
}

/*
 * @testcase app_control_uri_get_host_n
 * @description Gets a host subcomponent from a URI handle.
 *              The function returns a nullptr.
 * @apicovered app_control_uri_get_host
 */
TEST_F(UriTest, app_control_uri_get_host_n) {
  const char* host;
  int r = app_control_uri_get_host(nullptr, &host);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_get_port_p
 * @description Gets a port subcomponent from a URI handle.
 * @apicovered app_control_uri_get_port
 */
TEST_F(UriTest, app_control_uri_get_port_p) {
  const char* port;
  int r = app_control_uri_get_port(uri_, &port);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  EXPECT_NE(port, nullptr);
  EXPECT_EQ(std::string(port), "123");
}

/*
 * @testcase app_control_uri_get_port_n
 * @description Gets a port subcomponent from a URI handle.
 *              The function returns a nullptr.
 * @apicovered app_control_uri_get_port
 */
TEST_F(UriTest, app_control_uri_get_port_n) {
  const char* port;
  int r = app_control_uri_get_port(nullptr, &port);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_get_user_p
 * @description Gets a user subcomponent from a URI handle.
 * @apicovered app_control_uri_get_user
 */
TEST_F(UriTest, app_control_uri_get_user_p) {
  const char* user;
  int r = app_control_uri_get_user(uri_, &user);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  EXPECT_NE(user, nullptr);
  EXPECT_EQ(std::string(user), "john.doe");
}

/*
 * @testcase app_control_uri_get_user_n
 * @description Gets a user subcomponent from a URI handle.
 *              The function returns a nullptr.
 * @apicovered app_control_uri_get_user
 */
TEST_F(UriTest, app_control_uri_get_user_n) {
  const char* user;
  int r = app_control_uri_get_user(nullptr, &user);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
}

/*
 * @testcase app_control_uri_query_foreach_p
 * @description Retrieves the key-value pairs attributes in the query component.
 * @apicovered app_control_uri_query_foreach
 */
TEST_F(UriTest, app_control_uri_query_foreach_p) {
  // Pre-condition: Gets the query handle
  app_control_uri_query_h query = nullptr;
  int r = app_control_uri_get_query(uri_, &query);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  EXPECT_NE(query, nullptr);

  r = app_control_uri_query_foreach(query,
      [](const char* key, const char* val, void* user_data) -> bool {
        UriTest* p = static_cast<UriTest*>(user_data);
        if (!strcmp(key, "tag") && !strcmp(val, "networking"))
          p->count_++;
        else if (!strcmp(key, "order") && !strcmp(val, "newest"))
          p->count_++;
        return true;
      }, this);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  EXPECT_EQ(count_, 2);
}

/*
 * @testcase app_control_uri_query_foreach_n
 * @description Retrieves the key-value pairs attributes in the query component.
 *              The function returns a negative error value.
 * @apicovered app_control_uri_query_foreach
 */
TEST_F(UriTest, app_control_uri_query_foreach_n) {
  // Pre-condition: Gets the query handle
  app_control_uri_query_h query = nullptr;
  int r = app_control_uri_get_query(uri_, &query);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_NONE);
  EXPECT_NE(query, nullptr);

  r = app_control_uri_query_foreach(query, nullptr, this);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);

  r = app_control_uri_query_foreach(nullptr,
      [](const char* key, const char* val, void* user_data) -> bool {
        UriTest* p = static_cast<UriTest*>(user_data);
        if (!strcmp(key, "tag") && !strcmp(val, "networking"))
          p->count_++;
        else if (!strcmp(key, "order") && !strcmp(val, "newest"))
          p->count_++;
        return true;
      }, this);
  EXPECT_EQ(r, APP_CONTROL_URI_ERROR_INVALID_PARAMETER);
  EXPECT_EQ(count_, 0);
}
